#include <wgd/game.h>
#include <wgd/scene.h>

#include "gear.h"

using namespace wgd;

Gear::Gear(const doste::OID &id) : IModel(id) {
	mode(AABB);
	solid(true);
	//so collision knows what it is
	set("type", "platform"); 
}

//this must be called after the model is loaded
void Gear::initialise(){
	//get yaw from plane
	switch(plane()){
	case 0: set(ix::yaw, M_PI); 
		break;
		
	case 1: set(ix::yaw, M_PI_2);
		m_max = Vector3D(m_max.z, m_max.y, m_max.x);
		m_min = Vector3D(m_min.z, m_min.y, m_min.x);
		break;
		
	case 2: set(ix::yaw, M_PI);
		break;
		
	case 3: set(ix::yaw, M_PI+M_PI_2);
		m_max = Vector3D(m_max.z, m_max.y, m_max.x);
		m_min = Vector3D(m_min.z, m_min.y, m_min.x);
		break;
	}
	
	//Initialise any movement that hasnt been specified
	if(get("speed")!=Null){
		
		if(get("startx") == Null) set("startx", get(ix::x));
		if(get("starty") == Null) set("starty", get(ix::y));
		if(get("startz") == Null) set("startz", get(ix::z));
		if(get("endx") == Null) set("endx", get("startx"));
		if(get("endy") == Null) set("endy", get("starty"));
		if(get("endz") == Null) set("endz", get("startz"));
		
		position(Vector3D((float)get("startx"), (float)get("starty"), (float)get("startz")));
		
		
	}
	
}


void Gear::draw(){
	if(!Scene::current()->debug()) IModel::draw();
	
	glDisable(GL_LIGHTING);
	
	//Draw gear with lines
	if(Scene::current()->debug()){
		Vector2D p[6];
		
		glBegin(GL_LINE_LOOP);
		glColor3f(1.0, 1.0, 1.0);
		for(int i=0; i<teeth()*2; i+=2){
			seg(i  , p[0], p[1]);
			seg(i+1, p[2], p[3]);
			seg(i+2, p[4], p[5]);
			
			if(plane() == 0 || plane() == 2){
				for(int j=0; j<5; j++) glVertex3f(p[j].x, p[j].y, position().z);
			} else {
				for(int j=0; j<5; j++) glVertex3f(position().x, p[j].y, p[j].x);
			}
		}	
		glEnd();
			
		//Debug line
		if(test.length()!=0){
			glBegin(GL_LINES);
				glColor3f(1.0, 1.0, 0.0);
				glVertex3f(test.x, test.y, test.z);
				glVertex3f(test2.x, test2.y, test2.z);
			glEnd();	
		}
	}
	
	//Line to pivot point
	if(offset()!=0){
		glBegin(GL_LINES);
			glColor3f(0.8, 0.8, 0.8);
			glVertex3f(position().x, position().y, position().z);
			glVertex3f(pivot.x, pivot.y, pivot.z);
		glEnd();
	}
	
	//Line between start and end
	if(speed()!=0){
		glBegin(GL_LINES);
			glColor3f(0.3, 0.3, 0.3);
			glVertex3f((float)get("startx"), (float)get("starty"), (float)get("startz"));
			glVertex3f((float)get("endx"),   (float)get("endy"),   (float)get("endz"));
		glEnd();		
	}
	glEnable(GL_LIGHTING);
		
}

void Gear::update(){
	//acceleration for swinging
	float acc = -(float)get(ix::roll) * Game::frameTime() * force();
	rvelocity(rvelocity() + acc);
	
	//New angle
	float ang = rvelocity() * Game::frameTime() * (force()==0? 1: force());
	
	//offset
	if(offset()!=0){
		Vector2D off = shift(centre2D(), offset(), rotation() + M_PI);
		Vector2D off2 = shift(centre2D(), offset(), rotation() + ang + M_PI);
		pivot = off2;
		
		//move offset
		move(off-off2);
	}
	
	//rotate
	rotate(Vector3D(0, 0, ang));
	
	//move platform
	if(speed()!=0){
		//get movement vector
		Vector3D start((float)get("startx"), (float)get("starty"), (float)get("startz"));
		Vector3D end  ((float)get("endx"),   (float)get("endy"),   (float)get("endz"));
		Vector3D mv(end-start);
		
		move(mv * (1/speed()) * Game::frameTime());
		
		//if got to the end, go the other way
		if(speed()>0){
			Vector3D a = mv.normalise();
			Vector3D b = (position()-end).normalise();
			
			if(a.dot(b) > 0) {
				speed(-speed());
				rvelocity(-rvelocity());
			}
		} else {
			Vector3D a = mv.normalise();
			Vector3D b = (position()-start).normalise();
			
			if(a.dot(b) < 0){
				speed(-speed());
				rvelocity(-rvelocity());
			}
		}
	}
}

Vector2D Gear::velocityAt(const Vector3D &pos){
	Vector2D v;
//Actually methinks its only movement that needs to apply its velocity to the player

	//rotational
	Vector2D ctr = shift(centre2D(), offset(), rotation()+M_PI);
	//direction
	Vector2D dir = (pos - ctr).normalise();
	v.x = -dir.y; v.y = dir.x;
	//amplitude
	float amp =  (pos-ctr).length() * tan(rvelocity());
	v = v * amp;
	
	
	//directional
	if(speed()!=0){
		Vector3D start((float)get("startx"), (float)get("starty"), (float)get("startz"));
		Vector3D end  ((float)get("endx"),   (float)get("endy"),   (float)get("endz"));
		Vector3D mv(end-start);
		mv = mv * (1/speed());
		
		v = mv;
	}
	
	
	//debug line
	test = pos + v;
	test2 = pos;
	
	
	
	return v;
}

Vector2D Gear::centre2D(){
	Vector3D p = position();
	switch(plane()){
	case 0: 
	case 2: return Vector2D(p.x, p.y);
	case 1:
	case 3:	return Vector2D(p.z, p.y);
	}
}

Vector2D Gear::playerPos(const double &angle, const double &distance){
	return shift(centre2D(), distance, angle+rotation());
}

bool Gear::intersect(const Vector2D &pos, Vector2D &npos, Vector2D &tangent){
	//test if npos is inside, if so, set it to the intersection
	//assume pos is outside
	Vector2D ctr = centre2D();
	
	//get the 3 lines to test with (4 Vector2Ds)
	float cAng = atan2(npos.y-ctr.y, npos.x-ctr.x) + M_PI_2;
	float rAng = cAng - rotation();
	
	float step = M_PI/teeth();
	
	int tooth = (int)floor(rAng/step);
	
	Vector2D p[4];
	
	//which side?
	if(rAng/step-tooth < 0.5){
		seg(tooth-1, p[0], p[1]);
		seg(tooth,   p[2], p[3]);
	} else {
		seg(tooth,   p[0], p[1]);
		seg(tooth+1, p[2], p[3]);		
	}
	
	//do tangential lines first, then radial line
	bool a1 = above(npos, p[0], p[1]);
	bool a2 = above(npos, p[2], p[3]);
	bool a3 = above(npos, p[1], p[2]);
	
	
	if(a1&&a3 || a2&&a3 || a1&&a2){
		return false;
	} else {
		//cout << "Inside\n";
		Vector2D np;
		//get the intersection Vector2D
		if(!a1 && cross(np, pos, npos, p[0],p[1])) {
			npos = np;
			tangent = (p[1] - p[0]).normalise();
			return true;
		}
		if(!a2 && cross(np, pos, npos, p[2],p[3])){
			npos = np;
			tangent = (p[3] - p[2]).normalise();	
			return true;
		}
		if(!a3 && cross(np, pos, npos, p[1],p[2])){
			npos = np;
			tangent = (p[2] - p[1]).normalise();	
			return true;
		}
		
		//no collision - special case
		return false;
	}
}

bool Gear::inside(const Vector2D &pos){
	//Gear position and angle
	Vector2D ctr = centre2D();
	float cAng = atan2(pos.y-ctr.y, pos.x-ctr.x) + M_PI_2;
	float rAng = cAng - rotation();
	
	//Get the nearest tooth
	float step = M_PI/teeth();
	int tooth = (int)floor(rAng/step);
	
	Vector2D p[4];
	//which side - as only need 3 lines
	if(rAng/step-tooth < 0.5){
		seg(tooth-1, p[0], p[1]);
		seg(tooth,   p[2], p[3]);
	} else {
		seg(tooth,   p[0], p[1]);
		seg(tooth+1, p[2], p[3]);		
	}
	
	//do tangential lines first, then radial line
	bool a1 = above(pos, p[0], p[1]);
	bool a2 = above(pos, p[2], p[3]);
	bool a3 = above(pos, p[1], p[2]);
	
	return (a1&&a3 || a2&&a3 || a1&&a2);	
}

Vector2D Gear::shift(const Vector2D &pos, float dist, float angle) const {
	float tx = pos.x + dist * cos(angle-M_PI_2);
	float ty = pos.y + dist * sin(angle-M_PI_2);
	return Vector2D(tx, ty);
}

bool Gear::cross(Vector2D &pos, const Vector2D &a1, const Vector2D &a2, const Vector2D &b1, const Vector2D &b2){
	//Note: they share the same denominator
	float ua = (b2.x - b1.x) * (a1.y - b1.y) - (b2.y - b1.y) * (a1.x - b1.x);
	float ub = (a2.x - a1.x) * (a1.y - b1.y) - (a2.y - a1.y) * (a1.x - b1.x);
	float dn = (b2.y - b1.y) * (a2.x - a1.x) - (b2.x - b1.x) * (a2.y - a1.y);
	
	//lines are parallel, return coincident
	if(dn==0) return (ua==0 && ub==0); 
	
	//divide
	ua = ua / dn;
	ub = ub / dn;
	
	//get Vector2D where lines cross
	pos.x = a1.x + ua * (a2.x - a1.x);
	pos.y = a1.y + ua * (a2.y - a1.y);
	
	//do they cross?
	if(ua >= 0.0 && ua <= 1.0 && ub >= 0.0 && ub <= 1.0) return true;
	else return false;
}

bool Gear::above(const Vector2D &pos, const Vector2D &a1, const Vector2D &a2){
	
	Vector2D v1, v2;
	
	v1.x =  a2.y-a1.y;
	v1.y = -(a2.x-a1.x);
	v1.normalise();
	
	v2.x = pos.x - a1.x;
	v2.y = pos.y - a1.y;
	v2.normalise();
	
	return (v1.x*v2.x + v1.y*v2.y) >= 0;
}

void Gear::seg(int tooth, Vector2D &a, Vector2D &b){
	Vector2D t1, t2;
	Vector2D ctr = centre2D();
	float frotation = rotation();
	float fradius = radius();
	
	float size = M_PI / teeth();
	
	if(tooth%2==0){
		//outside
		float fbite = bite();
		float ftaper = taper();
		
		t1 = shift(ctr, fradius+fbite, tooth*size + frotation);
		t2 = shift(ctr, fradius+fbite, (tooth+1)*size + frotation);		
		
		a = Vector2D::lerp(t1, t2, ftaper/2);
		b = Vector2D::lerp(t2, t1, ftaper/2);		
		
	} else {
		//inside
		a = shift(ctr, fradius, tooth*size + frotation);
		b = shift(ctr, fradius, (tooth+1)*size + frotation);	
	}	
}
