/*
 * Using the Warwick Game Design C++ Library
 */

#include <wgd/wgd.h>
#include <wgd/scene.h>
#include <wgd/ilight.h>
#include "gear.h"
#include "player.h"

using namespace wgd;

class MyGame : public wgd::Game {
	public:
	MyGame(char *l) : Game(l){};
	
	//camera control
	Vector3D camTarget;
	float yawTarget;	
	
	bool initialise() {
		Scene::regInstance<Gear>();
		
		//load models
		DB::runScript("./scripts/models.dst");
		
		//load level1
		Scene::load("level1", "./scripts/level1.dst");
				
		Scene::current("level1");
		//Scene::current()->debug(true);
		start();
		
		Lighting::enabled(true);
		Instance::load<ILight>("light0", "./scripts/light.dst");
		
		//shortcuts
		WGD.set("g", WGD[ix::scenes][ix::cur][ix::instances]["gears"]);
		WGD.set("p", WGD[ix::scenes][ix::cur][ix::instances]["player"]["player"]);
		
		//paused
		WGD.set("menu", true);
		
		Camera::camera()->position(Vector3D(0, 0, 5));
		
		Instance::get<Player>("player")->update();
		
		camTarget.z = 5;
	};
	
	//loop through all gears calling initialise
	void start(){
		for(OID i=WGD[ix::scenes][ix::cur][ix::instances]["gears"].First(); i!=Null; i=WGD[ix::scenes][ix::cur][ix::instances]["gears"].Next()){
			if(i==Parent) continue;
			Instance::get<Gear>(i)->initialise();
			
		}
	}
	
	//Main update loop - all game logic and workings go in this function
	// do not draw stuff in this function
	bool update() {
		Scene::current()->update();
		
		if (Keyboard::keyPressed(Keyboard::KEY_SPACE))	WGD.set("menu", false);;
		
		if((bool)WGD["menu"] == false){
			Player *p = Instance::get<Player>("player");
			p->update();
			
			//camera - get lookat point
			if(p->cPlatform()!=NULL){
				Matrix mat;
				//mat.rotate(0, -p->cPlatform()->angle().y, 0);
				mat.rotate(0, -p->plane() * M_PI_2, 0);
				camTarget = p->cPlatform()->centre() + (mat * Vector3D(0,0,-5));
				
				yawTarget = p->plane() * M_PI_2 - M_PI;
			} 
			//move camera
			Vector3D target = Vector3D::lerp(Camera::camera()->position(), camTarget, 2 * frameTime());
			Camera::camera()->position(target);
			
			//rotate camera - note: have to wrap this...
			float cYaw = Camera::camera()->angle().y;
			float yaw = cYaw + (yawTarget-cYaw) *  (2 * frameTime());
			Camera::camera()->angle(Vector3D(0, yaw, 0));			
		
		}
		
		
		//If escape is pressed then quit the game.
		if (Keyboard::keyPressed(Keyboard::KEY_ESCAPE))	quit();
	};

	//Main draw loop, This function is where you draw everything in the game.
	bool draw(){
		Scene::current()->draw();
		
		Instance::get<Player>("player")->draw();
	}
	
	
	//This function gets called when the game exits.
	bool finalise() {
	}
};

int main() {
	MyGame *g = new MyGame("./scripts/config.dst");
	g->run();
	delete g;
}
